/*
 * Copyright (c) 2009, Jonathan L. Woodring and Han-Wei Shen
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *     * Redistributions of source code must retain the above copyright
 *       notice, this list of conditions and the following disclaimer.
 *     * Redistributions in binary form must reproduce the above copyright
 *       notice, this list of conditions and the following disclaimer in the
 *       documentation and/or other materials provided with the distribution.
 *     * Neither the name of The Ohio State University, Jonathan
 *       L. Woodring, Han-Wei Shen, nor the names of its contributors
 *       may be used to endorse or promote products derived from this 
 *       software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY JONATHAN L. WOODRING AND HAN-WEI SHEN
 * ''AS IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT 
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
 * FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE OHIO
 * STATE UNIVERSITY, JONATHAN L. WOODRING, OR HAN-WEI SHEN BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF 
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR
 * BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 * LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING 
 * NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#ifndef __ALLOCARRAY_H__
#define __ALLOCARRAY_H__

#include <string.h>

template <class Item>
class AllocArray;

template <class Item>
class AllocArray {
 public:
  AllocArray();
  ~AllocArray();

  Item& operator [](int i);
  int Size() const;
  void SetSize(int s);
  Item* Vector() const;
  void MemoryClear();
  void operator &=(AllocArray &o);

 protected:
  Item* array;
  int size;
};

template <class Item>
inline AllocArray<Item>::AllocArray() {
  size = 0;
  array = 0;
}

template <class Item>
inline AllocArray<Item>::~AllocArray() {
  delete [] array;
}

template <class Item>
inline Item& AllocArray<Item>::operator [](int i) {
  return array[i];
}

template <class Item>
inline int AllocArray<Item>::Size() const {
  return size;
}

template <class Item>
inline void AllocArray<Item>::SetSize(int s) {
  delete [] array;
  array = new Item[s];
  size = s;
}

template <class Item>
inline Item* AllocArray<Item>::Vector() const {
  return array;
}

template <class Item>
inline void AllocArray<Item>::MemoryClear() {
  memset((void*) array, 0, sizeof(Item)*size);
}

template <class Item>
inline void AllocArray<Item>::operator &= (AllocArray &o) {
  Item* temparray;
  int tempsize;
  
  temparray = array;
  tempsize = size;

  size = o.size;
  array = o.array;

  o.size = tempsize;
  o.array = temparray;
}

#endif







