/*
 * Copyright (c) 2009, Jonathan L. Woodring and Han-Wei Shen
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *     * Redistributions of source code must retain the above copyright
 *       notice, this list of conditions and the following disclaimer.
 *     * Redistributions in binary form must reproduce the above copyright
 *       notice, this list of conditions and the following disclaimer in the
 *       documentation and/or other materials provided with the distribution.
 *     * Neither the name of The Ohio State University, Jonathan
 *       L. Woodring, Han-Wei Shen, nor the names of its contributors
 *       may be used to endorse or promote products derived from this 
 *       software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY JONATHAN L. WOODRING AND HAN-WEI SHEN
 * ''AS IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT 
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
 * FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE OHIO
 * STATE UNIVERSITY, JONATHAN L. WOODRING, OR HAN-WEI SHEN BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF 
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR
 * BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 * LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING 
 * NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#ifndef __SEQUENTIAL_SPACE_H__
#define __SEQUENTIAL_SPACE_H__

#include "Mapper.h"
#include "AllocArray.h"

class SequentialSpaceReader
{
 public:
  SequentialSpaceReader(int mmap, char* format, 
			unsigned int start, 
			unsigned int end, 
			unsigned int skip);
  ~SequentialSpaceReader();

  float getFloat(unsigned int coord);
  void getFloat(float* buffer, unsigned int coord, unsigned int count);

  void setTime(unsigned int time);

  unsigned int getSize();
  unsigned int getMaxTime();
  unsigned int getTime();
  AllocArray<unsigned int>* getDims();

 private:
  unsigned int tc;
  Mapper** mappers;
  unsigned int current;
  unsigned int vc;

  AllocArray<unsigned int> dims;
};

class SequentialSpaceWriter
{
 public:
  SequentialSpaceWriter(char* format, char* meta, char* data,
			unsigned int tc, AllocArray<unsigned int>& oridims,
			unsigned int offset, unsigned int skip);
  ~SequentialSpaceWriter();

  void writeTimeVector(float* buffer);
  void close();

 private:
  unsigned int tc;
  unsigned int offset;
  unsigned int skip;
  
  char* metafn;
  char* datafn;

  FILE** outputs;
  AllocArray<unsigned int> dims;
};

#endif
