/*
 * Copyright (c) 2009, Jonathan L. Woodring and Han-Wei Shen
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *     * Redistributions of source code must retain the above copyright
 *       notice, this list of conditions and the following disclaimer.
 *     * Redistributions in binary form must reproduce the above copyright
 *       notice, this list of conditions and the following disclaimer in the
 *       documentation and/or other materials provided with the distribution.
 *     * Neither the name of The Ohio State University, Jonathan
 *       L. Woodring, Han-Wei Shen, nor the names of its contributors
 *       may be used to endorse or promote products derived from this 
 *       software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY JONATHAN L. WOODRING AND HAN-WEI SHEN
 * ''AS IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT 
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
 * FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE OHIO
 * STATE UNIVERSITY, JONATHAN L. WOODRING, OR HAN-WEI SHEN BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF 
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR
 * BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 * LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING 
 * NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#ifndef __CLUSTER_OPS_H__
#define __CLUSTER_OPS_H__ 1

#include "SequentialTime.h"

#define DC_VECTOR 0
#define DC_DIFF_VECTOR 1
#define DIFF_VECTOR 2
#define HAAR_DC_VECTOR 3
#define HAAR_DC_DIFF_VECTOR 4
#define HAAR_DIFF_VECTOR 5

#define KMEANS_CLUSTERING 0

#define BOX_FILTER 0

#define EUCLIDEAN_DISTANCE 0


double distL2(float* a, double* b, unsigned int w);

//double distDTW2(float* a, float* b, unsigned int w);

void sum(double* sum, float *a, unsigned int w);

void clearBuffer(double* sums, float* errors, unsigned int* counts,
		 unsigned int w, unsigned int k);

void findBMUL2(float* buffer, double* sums,
	       unsigned int &bmu, double &mindist,
	       unsigned int w, unsigned int k);

/*
void findBMUDTW2(float* buffer, float* sums,
		 unsigned int &bmu, double &mindist,
		 unsigned int w, unsigned int k);
*/

void averageSums(double* sums, unsigned int* counts,
		 unsigned int w, unsigned int k);

float averageErrors(float* errors, unsigned int* counts, unsigned int k);

unsigned int jenkins(double* sums, unsigned int* counts,
		     unsigned int w, unsigned int k);

void calcDifferences(float* buffer, unsigned int w);

void calcDifferences(float* buffer, float* dest, unsigned int b);

void getTimeVectorBox(SequentialTimeReader* input, float* buffer,
		      unsigned int coord, unsigned int l, unsigned int r);

void getTimeVecAndDiffBox(SequentialTimeReader* input, float* buffer,
			  unsigned int coord, unsigned int l, unsigned int r);

void getTimeDiffBox(SequentialTimeReader* input, float* buffer,
		    unsigned int coord, unsigned int l, unsigned int r);

void learn(float* sum, float* buffer, unsigned int w, float rate);

void writeInfo(char* format, char* infoExt, unsigned int t,
	       double* sums, float* errors, unsigned int* counts,
	       unsigned int* histograms,
	       unsigned int clusteringType,
	       unsigned int vectorType,
	       unsigned int filterType,
	       unsigned int distanceType,
	       unsigned int w, unsigned int k, unsigned int bins,
	       float min, float max, unsigned int iter, float aveerr);

void calcHaarDC(float* buffer, unsigned int b);

void calcHaarDCDiff(float* buffer, unsigned int b);

void calcHaarDiff(float* buffer, float* dest, unsigned int b);

void getTimeHaarDCBox(SequentialTimeReader* input, float* buffer,
		      unsigned int coord, unsigned int l, unsigned int r);

void getTimeHaarDCDiffBox(SequentialTimeReader* input, float* buffer,
			  unsigned int coord, unsigned int l, unsigned int r);

void getTimeHaarDiffBox(SequentialTimeReader* input, float* buffer,
			unsigned int coord, unsigned int l, unsigned int r);

#endif

