/*
 * Copyright (c) 2009, Jonathan L. Woodring and Han-Wei Shen
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *     * Redistributions of source code must retain the above copyright
 *       notice, this list of conditions and the following disclaimer.
 *     * Redistributions in binary form must reproduce the above copyright
 *       notice, this list of conditions and the following disclaimer in the
 *       documentation and/or other materials provided with the distribution.
 *     * Neither the name of The Ohio State University, Jonathan
 *       L. Woodring, Han-Wei Shen, nor the names of its contributors
 *       may be used to endorse or promote products derived from this 
 *       software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY JONATHAN L. WOODRING AND HAN-WEI SHEN
 * ''AS IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT 
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
 * FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE OHIO
 * STATE UNIVERSITY, JONATHAN L. WOODRING, OR HAN-WEI SHEN BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF 
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR
 * BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 * LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING 
 * NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include "psplitCommand.h"
#include <stdio.h>
#include <stdlib.h>

#include "SequentialTime.h"

#include <mpi.h>
#include "mpiExterns.h"

int main(int argc, char* argv[])
{
  MPI_Init(&argc, &argv);
  MPI_Comm_size(MPI_COMM_WORLD, &mpi_size);
  MPI_Comm_rank(MPI_COMM_WORLD, &mpi_rank);

  CommandArguments* args = commandParse(argc, argv);

  SequentialTimeReader* input =
    new SequentialTimeReader(args->useMMap, args->inputFormat);

  if(args->startFile >= args->endFile) {
    args->endFile = input->getMaxTime();
  }

  unsigned int r = args->endFile;
  unsigned int l = args->startFile;
  unsigned int s = args->skipFile;

  unsigned int chunk = input->getSize() / mpi_size;
  unsigned int rem = input->getSize() % mpi_size;

  unsigned int sv = chunk * mpi_rank + (mpi_rank < rem ? mpi_rank : rem); 
  unsigned int ev = chunk * (mpi_rank + 1) +
    (mpi_rank + 1 < rem ? mpi_rank + 1 : rem);

  char metaformat[strlen(args->metaExt) + 256];
  sprintf(metaformat, "%s[%d]", args->metaExt, mpi_rank); 
  char dataformat[strlen(args->dataExt) + 256];
  sprintf(dataformat, "%s[%d]", args->dataExt, mpi_rank); 

  SequentialTimeWriter* output =
    new SequentialTimeWriter(args->outputFormat, metaformat, dataformat,
			     r - l, args->maxFile, *input->getOriginalDims(),
			     ev - sv);

  // read your the vectors and write to your portion
  float buffer[(r - l) / s];
  if(s > 1) {
    for(unsigned int i = sv; i < ev; i = i + 1) {
      input->getTimeVector(buffer, i, l, r, s);
      output->writeTimeVector(buffer);
    }
  }
  else {
    if(l == 0 && r == input->getMaxTime()) {
      for(unsigned int i = sv; i < ev; i = i + 1) {
	input->getTimeVector(buffer, i);
	output->writeTimeVector(buffer);
      }
    }
    else {
      for(unsigned int i = sv; i < ev; i = i + 1) {
	input->getTimeVector(buffer, i, l, r);
	output->writeTimeVector(buffer);
      }
    }
  }

  output->setMin(input->getMin());
  output->setMax(input->getMax());
  
  delete input;
  delete output;
  delete args;

  MPI_Finalize();
}
