/*
 * Copyright (c) 2009, Jonathan L. Woodring and Han-Wei Shen
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *     * Redistributions of source code must retain the above copyright
 *       notice, this list of conditions and the following disclaimer.
 *     * Redistributions in binary form must reproduce the above copyright
 *       notice, this list of conditions and the following disclaimer in the
 *       documentation and/or other materials provided with the distribution.
 *     * Neither the name of The Ohio State University, Jonathan
 *       L. Woodring, Han-Wei Shen, nor the names of its contributors
 *       may be used to endorse or promote products derived from this 
 *       software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY JONATHAN L. WOODRING AND HAN-WEI SHEN
 * ''AS IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT 
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
 * FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE OHIO
 * STATE UNIVERSITY, JONATHAN L. WOODRING, OR HAN-WEI SHEN BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF 
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR
 * BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 * LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING 
 * NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include "ritCommand.h"
#include <stdio.h>
#include <stdlib.h>

#include "SequentialSpace.h"
#include "SequentialTime.h"

int main(int argc, char* argv[])
{
  CommandArguments* args = commandParse(argc, argv);

  if(args->startFile >= args->endFile) {
    args->endFile = args->startFile + 1;
  }

  if(args->maxFile < (args->endFile - args->startFile) * sizeof(float)) {
    args->maxFile = (args->endFile - args->startFile) * sizeof(float);
  }

  SequentialSpaceReader input(args->useMMap, 
			      args->inputFormat, args->startFile,
			      args->endFile, args->skipFile);

  unsigned int chunk = args->chunkSize;
  unsigned int vc = input.getSize();
  unsigned int tc = input.getMaxTime();
  float* buffer = new float[chunk * tc];
  
  SequentialTimeWriter output(args->outputFormat,
			      args->metaExt, args->dataExt,
			      input.getMaxTime(), args->maxFile,
			      *input.getDims(), vc);

  float min = input.getFloat(0);
  float max = min;

  // for all the spatial points
  for(unsigned int i = 0; i < vc; i = i + chunk) {
    if(i + chunk > vc) {
      chunk = vc - i;
    }

    // read in chunk size vectors
    for(unsigned int j = 0; j < tc; j = j + 1) {

      input.setTime(j);
      for(unsigned int k = 0; k < chunk; k = k + 1) {
	buffer[k * tc + j] = input.getFloat(i + k);
	
	min = min < buffer[k * tc + j] ? min : buffer[k * tc + j];
	max = max > buffer[k * tc + j] ? max : buffer[k * tc + j];
      }
    }

    // write it out to the output file
    for(unsigned int j = 0; j < chunk; j = j + 1) {
      output.writeTimeVector(&(buffer[j * tc]));
    }
  }

  output.setMin(min);
  output.setMax(max);

  delete [] buffer;
  delete args;
}
