#!/usr/bin/python

#/*
# * Copyright (c) 2009, Jonathan L. Woodring and Han-Wei Shen
# * All rights reserved.
# *
# * Redistribution and use in source and binary forms, with or without
# * modification, are permitted provided that the following conditions are met:
# *     * Redistributions of source code must retain the above copyright
# *       notice, this list of conditions and the following disclaimer.
# *     * Redistributions in binary form must reproduce the above copyright
# *       notice, this list of conditions and the following disclaimer in the
# *       documentation and/or other materials provided with the distribution.
# *     * Neither the name of The Ohio State University, Jonathan
# *       L. Woodring, Han-Wei Shen, nor the names of its contributors
# *       may be used to endorse or promote products derived from this
# *       software without specific prior written permission.
# *
# * THIS SOFTWARE IS PROVIDED BY JONATHAN L. WOODRING AND HAN-WEI SHEN
# * ''AS IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
# * FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE OHIO
# * STATE UNIVERSITY, JONATHAN L. WOODRING, OR HAN-WEI SHEN BE LIABLE
# * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR
# * BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
# * LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
# * NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
# */

from os import *
from sys import *
from re import *
from math import *

#
# dummy function
#
class distFunc:
    def __init__(self):
        self.dist = self.EMDdist

    #
    # chi^2 histogram distance
    #
    def chi2dist(self, a, b):
        ta = reduce(lambda x, y: x + y, a)
        tb = reduce(lambda x, y: x + y, b)
        tm = max(ta, tb)

        a = map(lambda x: x * tm / float(ta), a)
        b = map(lambda x: x * tm / float(tb), b)

        sum = 0.0
        for i, j in zip(a, b):
            if i + j > 0:
                sum = sum + (i - j) * (i - j) / float(i + j)

        if sum < 1:
            return .5

        return sum

    #
    # L2 histogram distance
    #
    def L2dist(self, a, b):
        ta = reduce(lambda x, y: x + y, a)
        tb = reduce(lambda x, y: x + y, b)
        tm = max(ta, tb)

        a = map(lambda x: x * tm / float(ta), a)
        b = map(lambda x: x * tm / float(tb), b)

        result = reduce(lambda x, y: x + (y[0] - y[1]) * (y[0] - y[1]), zip(a, b), 0.0)
        if result < 1:
            return .5

        return result

    #
    # EMD 1-D scan distance
    #
    def EMDdist(self, a, b):
        ta = reduce(lambda x, y: x + y, a)
        tb = reduce(lambda x, y: x + y, b)
        tm = max(ta, tb)

        a = map(lambda x: x * tm / float(ta), a)
        b = map(lambda x: x * tm / float(tb), b)

        moved = 0

        # move earth around by pushing or pulling from the right,
        # left will always be correct
        # ignore the last bin (overflow)
        for i in range(0, len(a) - 1):
            if a[i] > b[i]:
                j = i + 1
                while a[i] > b[i] and j < len(b):
                    if a[j] < b[j]:
                        c = a[i] - b[i]
                        d = b[j] - a[j]

                        if d > c:
                            moved = moved + c * (j - i)
                            a[j] = a[j] + c
                            a[i] = a[i] - c
                        else:
                            moved = moved + d * (j - i)
                            a[j] = a[j] + d
                            a[i] = a[i] - d

                    j = j + 1


            elif a[i] < b[i]:
                j = i + 1
                while a[i] < b[i] and j < len(b):
                    if a[j] > b[j]:
                        c = b[i] - a[i]
                        d = a[j] - b[j]

                        if d > c:
                            moved = moved + c * (j - i)
                            a[j] = a[j] - c
                            a[i] = a[i] + c
                        else:
                            moved = moved + d * (j - i)
                            a[j] = a[j] - d
                            a[i] = a[i] + d

                    j = j + 1

        if moved < 1:
            return .5

        return moved                

DIST = distFunc()

#    
# find all paths in a graph
#
def findAllPaths(graph, start, end, path=[]):
    path = path + [start]
    if start == end:
        return [path]
    if not graph.has_key(start):
        return []
    paths = []
    for node in graph[start]:
        if node not in path:
            newpaths = findAllPaths(graph, node, end, path)
            for newpath in newpaths:
                paths.append(newpath)
    return paths

#
# adjust a single histogram
#

def adjustOneHistogram(histo, minv, maxv, rangev, cmin, cmax, crange, newbins):
    newhisto = [0] * newbins
    lenhisto = float(len(histo))
    lennewhisto = float(len(newhisto))

    for n in range(0, len(histo)):
        s = crange * n / lenhisto + cmin
        e = crange * (n + 1) / lenhisto + cmin
        sf, si = modf(lennewhisto * (s - minv) / rangev)
        ef, ei = modf(lennewhisto * (e - minv) / rangev)
        si = int(si)
        ei = int(ei)

        if ei >= lennewhisto:
            ei = int(lennewhisto - 1)
            ef = 1.0

        t = histo[n]
        tl = ei + ef - si - sf
        tr = t

        if si == ei:
            newhisto[si] = newhisto[si] + t
            tr = tr - t

        for x in range(si, ei):
            if x == ei:
                newhisto[x] = newhisto[x] + t * ef / tl
                tr = tr - t * ef / tl
            elif x == si:
                newhisto[x] = newhisto[x] + t * (1.0 - sf) / tl
                tr = tr - t * (1.0 - sf) / tl
            else:
                newhisto[x] = newhisto[x] + t / tl
                tr = tr - t / tl

        if tr > 0:
            newhisto[ei] = newhisto[ei] + tr
            tr = 0

    return newhisto

#
# adjust the displayed histograms to be in the same min-max range
#
def adjustDisplayedHistograms(header, timesteps, newbins):
    minv = min(map(lambda x: x['min'], timesteps))
    maxv = max(map(lambda x: x['max'], timesteps))
    rangev = float(maxv - minv)

    for i in range(0, len(timesteps)):
        cmin = timesteps[i]['min']
        cmax = timesteps[i]['max']
        crange = float(cmax - cmin)

        for j in range(0, len(timesteps[i]['clusters'])):
            timesteps[i]['clusters'][j]['o_timehistogram'] = []

            for k in range(0, len(timesteps[i]['clusters'][j]['timehistogram'])):
                histo = timesteps[i]['clusters'][j]['timehistogram'][k]
                newhisto = adjustOneHistogram(histo, minv, maxv, rangev, cmin, cmax, crange, newbins)

                timesteps[i]['clusters'][j]['timehistogram'][k] = newhisto
                timesteps[i]['clusters'][j]['o_timehistogram'].append(histo)

        #timesteps[i]['min'] = cmin
        #timesteps[i]['max'] = cmax

    header['o_bins'] = header['bins']
    header['bins'] = newbins
    header['min'] = minv
    header['max'] = maxv
    return header, timesteps

#
# generate edge transition probabilties
#
def generateProbs(timesteps, power, header):
    weights = {}
    graph = {}

    # calculate the forward transition probabilities
    for i in range(0, len(timesteps) - 1):
        for j in range(0, len(timesteps[i]['clusters'])):
            dists = []
            graph[(j, i)] = []
            for k in range(0, len(timesteps[i+1]['clusters'])):
                # create an edge in the graph
                graph[(j, i)].append((k, i + 1))
                # total histogram distance
                dist = 0.0
                for l in range(1, header['width']):
                    a = timesteps[i]['clusters'][j]['o_timehistogram'][l]
                    b = timesteps[i+1]['clusters'][k]['o_timehistogram'][l-1]
                    amin = timesteps[i]['min']
                    bmin = timesteps[i+1]['min']
                    amax = timesteps[i]['max']
                    bmax = timesteps[i+1]['max']
                    mmin = min(amin, bmin)
                    mmax = max(amax, bmax)
                    a = adjustOneHistogram(a, mmin, mmax, mmax - mmin,\
                                               amin, amax, amax - amin,\
                                               header['o_bins'])
                    b = adjustOneHistogram(b, mmin, mmax, mmax - mmin,\
                                               bmin, bmax, bmax - bmin,\
                                               header['o_bins'])

                    dist = dist + DIST.dist(a, b)
                # normalized by buckets
                dist = dist / (header['width'] * header['o_bins'])
                dists.append(dist)

            # (1 / distance) ^ power probabilities
            total = 1.0 / reduce(lambda x, y: x + pow(1.0 / y, power), dists, 0.0)
            for k in range(0, len(dists)):
                weights[(j, i, k)] = pow(1.0 / dists[k], power) * total
    # empty forward edges
    for j in range(0, len(timesteps[-1]['clusters'])):
        graph[(j, len(timesteps) - 1)] = []

    # calculate the backward transition probabilities
    for i in range(-1, -len(timesteps), -1):
        for j in range(0, len(timesteps[i]['clusters'])):
            dists = []
            for k in range(0, len(timesteps[i-1]['clusters'])):
                # total histogram distance
                dist = 0.0
                for l in range(0, header['width'] - 1):
                    a = timesteps[i]['clusters'][j]['o_timehistogram'][l]
                    b = timesteps[i-1]['clusters'][k]['o_timehistogram'][l+1]
                    amin = timesteps[i]['min']
                    bmin = timesteps[i-1]['min']
                    amax = timesteps[i]['max']
                    bmax = timesteps[i-1]['max']
                    mmin = min(amin, bmin)
                    mmax = max(amax, bmax)
                    a = adjustOneHistogram(a, mmin, mmax, mmax - mmin,\
                                               amin, amax, amax - amin,\
                                               header['o_bins'])
                    b = adjustOneHistogram(b, mmin, mmax, mmax - mmin,\
                                               bmin, bmax, bmax - bmin,\
                                               header['o_bins'])

                    dist = dist + DIST.dist(a, b)
                # normalized by buckets
                dist = dist / (header['width'] * header['o_bins'])
                dists.append(dist)

            # (1 / distance) ^ power probabilities
            total = 1.0 / reduce(lambda x, y: x + pow(1.0 / y, power), dists, 0.0)
            for k in range(0, len(dists)):
                weights[(j, i, k)] = pow(1.0 / dists[k], power) * total 

    return graph, weights


#
# cull edges
#
def cullEdges(graph, weights, cull, ts):

    # find the max weight for forward and backward edge
    for i in graph.keys():
        for j in graph[i]:
            mv = max(weights[(i[0], i[1], j[0])],\
                         weights[(i[0], -(ts - i[1] - 1), j[0])])
            weights[(i[0], i[1], j[0])] = mv
            del weights[(i[0], -(ts - i[1] - 1), j[0])]

    forward = {}
    toremove = []
    for i in graph.keys():
        for j in graph[i]:
            if i[1] >= 0:
                if weights[(i[0], i[1], j[0])] < cull:
                    toremove.append((i, j))
                else:
                    forward[j] = None

    for i in toremove:
        graph[i[0]].remove(i[1])

    # find starting, ending, isolated nodes
    starters = {}
    enders = {}
    isolated = {}
    for i in graph.keys():
        if i[1] >= 0:
            if i not in forward:
                if len(graph[i]) <= 0:
                    isolated[i] = None
                else:
                    starters[i] = None
            elif len(graph[i]) <= 0:
                enders[i] = None

    return starters.keys(), enders.keys(), isolated.keys()


#
# substring search
# using Boyer-Moore-Horspool
#
def findSubstring(p, t):
    m = len(p)
    n = len(t)
    if m > n:
        return -1
    skip = {}
    for k in range(0, m - 1):
        skip[p[k]] = m - k - 1
    k = m - 1
    while k < n:
        j = m - 1
        i = k
        while j >= 0 and t[i][0] == p[j][0] and t[i][1] == p[j][1]:
            j = j - 1
            i = i - 1
        if j == -1:
            return i + 1
        if t[k] in skip:
            k = k + skip[t[k]]
        else:
            k = k + m
    return -1


#
# find the sequences
#
def pathFind(graph, starters, enders, ts):
    paths = []
    for i in starters:
        for j in enders:
            for k in findAllPaths(graph, i, j):
                paths.append(k)
    return paths


#
# remove duplicate or substring paths
#
def removeDuplicates(paths):
    nondup = []
    dup = {}
    for i in range(0, len(paths)):
        if i not in dup:
            contain = i
            for j in range(i, len(paths)):
                if len(paths[i]) < len(paths[j]):
                    if findSubstring(paths[i], paths[j]) >= 0:
                        contain = j
                        break
                else:
                    if findSubstring(paths[j], paths[i]) >= 0:
                        dup[j] = None
            
            if contain == i:
                nondup.append(paths[i])

    return nondup


#
# calculate some metrics of the goodness of the paths
#
def calcMetrics(weights, paths, ts):
    allmetrics = []
    for i in paths:
        metrics = {}
        metrics['length'] = len(i)

        flist = []
        fsum = 0.0
        fcount = 0
        fmul = 1.0

        for j in range(0, len(i) - 1):
            e = (i[j][0], i[j][1], i[j+1][0])
            if e in weights:
                val = weights[e]
                flist.append(val)
                fsum = fsum + val
                fmul = fmul * val
                fcount = fcount + 1

        if fcount > 0:
            metrics['min'] = min(flist)
            metrics['avg'] = fsum / float(fcount)
            metrics['mul'] = fmul
            metrics['count'] = fcount
        metrics['start'] = i[0][1]
        metrics['end'] = i[-1][1]
        
        allmetrics.append(metrics)

    return allmetrics


#
# cdf of counts
#

def cdf(h):
    result = [0]

    total = 0
    for i in h:
        total = total + i
        result.append(total)

    return result

#
# normalized value to cummulative count
#

def nv2cc(v, cdf):
    f, a = modf(v * float(len(cdf) - 1))

    a = int(a)
    if a < 0:
        return 0.0
    if a >= len(cdf) - 1:
        return cdf[-1]
    b = a + 1

    # linear interpolation
    return cdf[a] * (1.0 - f) + cdf[b] * f

#
# cummulative count to normalized value
#

def cc2nv(p, cdf):
    if p <= 0:
        return 0.0
    if p >= cdf[-1]:
        return 1.0

    i = 0
    while i < len(cdf) - 2 and cdf[i] < p:
        i = i + 1

    f = (p - cdf[i])
    d = float(cdf[i + 1] - cdf[i])
    if d > 0:
        f = f / d
    else:
        f = 1.0
    a = i / float(len(cdf) - 1)
    b = (i + 1) / float(len(cdf) - 1)
    
    return a * (1.0 - f) + b * f

#
# uniform to box opacity
#

def boxOpacity(c, a):
    def decide(x):
        if x <= 0 or x >= c[-1]:
            return 0.0
        else:
            return a

    return map(decide, c)

#
# uniform to sine curve opacity
#

def sineOpacity(c, a, p=.2):
    def opac(v):
        return a * pow((sin(2 * pi * v - pi * .5) + 1) * .5, p)

    return map(opac, map(lambda x: x / float(c[-1]), c))

#
# hsv2rgb
#

def hsv2rgb(h, s, v):
    if h < 0:
        return [0.0, 0.0, 0.0]

    f, b = modf(h / 60.0)
    b = int(b)
    p = v * (1 - s)
    q = v * (1 - s * f)
    r = v * (1 - s * (1 - f))

    if b == 0:
        return [v, r, p]
    if b == 1:
        return [q, v, p]
    if b == 2:
        return [p, v, r]
    if b == 3:
        return [p, q, v]
    if b == 4:
        return [r, p, v]

    return [v, p, q]


