#!/usr/bin/python

#/*
# * Copyright (c) 2009, Jonathan L. Woodring and Han-Wei Shen
# * All rights reserved.
# *
# * Redistribution and use in source and binary forms, with or without
# * modification, are permitted provided that the following conditions are met:
# *     * Redistributions of source code must retain the above copyright
# *       notice, this list of conditions and the following disclaimer.
# *     * Redistributions in binary form must reproduce the above copyright
# *       notice, this list of conditions and the following disclaimer in the
# *       documentation and/or other materials provided with the distribution.
# *     * Neither the name of The Ohio State University, Jonathan
# *       L. Woodring, Han-Wei Shen, nor the names of its contributors
# *       may be used to endorse or promote products derived from this
# *       software without specific prior written permission.
# *
# * THIS SOFTWARE IS PROVIDED BY JONATHAN L. WOODRING AND HAN-WEI SHEN
# * ''AS IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
# * FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE OHIO
# * STATE UNIVERSITY, JONATHAN L. WOODRING, OR HAN-WEI SHEN BE LIABLE
# * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR
# * BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
# * LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
# * NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
# */

from os import *
from sys import *
from tempfile import *
from math import *
from Tkinter import *
from tkFileDialog import *
from cPickle import *
from gzip import *

from sequence_funcs import *

#
# the GUI for the cluster data
#
class GUI:
    def __init__(self, parent, data, header, timesteps, graph, weights, paths, metrics):
        self.PWIDTH = 5
        self.THWIDTH = self.PWIDTH * header['width']
        self.THHEIGHT = header['bins']
        self.THXGAP = 20
        self.THYGAP = 10

        # find the maximum number of clusters in a timestep
        self.maxclusters = max(map(lambda x: len(x['clusters']), timesteps))
        self.ts = len(timesteps)
        self.width = header['width']
        self.height = header['bins']
        self.o_bins = header['o_bins']

        # set up the GUI
        parent.grid_columnconfigure(index=0, weight=1)
        parent.grid_rowconfigure(index=0, weight=1)

        self.canvas = Canvas(parent, width=1024, height=768, scrollregion=(0, 0, self.ts * (self.THWIDTH + self.THXGAP) + self.THXGAP * 2 + self.ts * self.PWIDTH, (self.maxclusters + len(paths)) * (self.THHEIGHT + self.THYGAP) + 3 * self.THYGAP), bg='black')
        self.vscroll = Scrollbar(parent, orient=VERTICAL)
        self.hscroll = Scrollbar(parent, orient=HORIZONTAL)
        self.listbox = Listbox(parent, width=50, selectmode=MULTIPLE)
        self.lbvscroll = Scrollbar(parent, orient=VERTICAL)
        self.lbhscroll = Scrollbar(parent, orient=HORIZONTAL)

        self.bottomrow = Frame(parent)
        self.opacitylabel = Label(self.bottomrow, text='Opacity')
        self.opacityentry = Entry(self.bottomrow)
        self.opacityentry.insert(0, '.1')
        self.opowerlabel = Label(self.bottomrow, text='Opacity Power')
        self.opowerentry = Entry(self.bottomrow)
        self.opowerentry.insert(0, '.2')
        self.savehsvSS = Button(self.bottomrow,\
                                    text='Save HSV Static/Static',\
                                    command=self.saveHSVStaticStatic)
        self.savehsvDD = Button(self.bottomrow,\
                                    text='Save HSV Dynamic/Dynamic',\
                                    command=self.saveHSVDynamicDynamic)
        self.savehsvSD = Button(self.bottomrow,\
                                    text='Save HSV Static/Dynamic',\
                                    command=self.saveHSVStaticDynamic)
        self.savehsvDS = Button(self.bottomrow,\
                                    text='Save HSV Dynamic/Static',\
                                    command=self.saveHSVDynamicStatic)
        self.saveSStatic = Button(self.bottomrow,\
                                      text='Save Simple Static',\
                                      command=self.saveSimpleStatic)
        self.saveSDynamic = Button(self.bottomrow,\
                                       text='Save Simple Dynamic',\
                                       command=self.saveSimpleDynamic)

        self.canvas.config(yscrollcommand=self.vscroll.set)
        self.canvas.config(xscrollcommand=self.hscroll.set)
        self.vscroll.config(command=self.canvas.yview)
        self.hscroll.config(command=self.canvas.xview)
        self.listbox.config(yscrollcommand=self.lbvscroll.set)
        self.listbox.config(xscrollcommand=self.lbhscroll.set)
        self.lbvscroll.config(command=self.listbox.yview)
        self.lbhscroll.config(command=self.listbox.xview)

        # some strange bug where S gets set to 16 instead of 's'
        N = 'n'
        S = 's'
        W = 'w'
        E = 'e'

        self.vscroll.grid(row=0, column=1, sticky=N+S)
        self.hscroll.grid(row=1, column=0, sticky=E+W)
        self.listbox.grid(row=0, column=2, sticky=N+S)
        self.lbvscroll.grid(row=0, column=3, sticky=N+S)
        self.lbhscroll.grid(row=1, column=2, sticky=E+W)
        self.canvas.grid(row=0, column=0, sticky=N+S+E+W)

        self.bottomrow.grid(row=2, columnspan=3)
        self.opacitylabel.grid(row=0, column=0)
        self.opacityentry.grid(row=0, column=1)
        self.opowerlabel.grid(row=0, column=2)
        self.opowerentry.grid(row=0, column=3)
        self.savehsvSS.grid(row=1, column=0)
        self.savehsvDD.grid(row=1, column=1)
        self.savehsvSD.grid(row=1, column=2)
        self.savehsvDS.grid(row=1, column=3)
        self.saveSStatic.grid(row=2, column=1)
        self.saveSDynamic.grid(row=2, column=2)

        # poll the listbox
        self.lbselect = None
        self.pathselect = []
        self.nodeselect = []
        self.lbPoll()

        # store the data
        self.data = data
        self.header = header
        self.timesteps = timesteps
        self.graph = graph
        self.weights = weights
        self.paths = paths
        self.metrics = metrics

        # create time histogram images per cluster
        self.imageCreate()

        # create time histogram images per path
        self.pathImageCreate()

        # display the images
        self.imageDisplay()

        # create edges and listbox entries
        self.pathDisplay()

        # create single paths at the bottom
        self.singlePathDisplay()

    # accumulate the entire time sequence into one distribution
    def accumulateAll(self):
        total = [0] * 510

        min = header['min']
        max = header['max']

        values = {}
        # for all the paths
        for i in self.pathselect:
            # create a set of clusters
            for j in self.paths[i]:
                values[j] = None

        # add the clusters up
        for j in values.keys():
            c = self.timesteps[j[1]]['clusters'][j[0]]['o_timehistogram'][self.width / 2]
            cmin = self.timesteps[j[1]]['min']
            cmax = self.timesteps[j[1]]['max']

            c = adjustOneHistogram(c, min, max, max - min,\
                                       cmin, cmax, cmax - cmin, 510)

            total = map(lambda x: x[0] + x[1], zip(total, c))

        return total

    # accumulate all the clusters for one time step
    def accumulate(self, t):
        total = [0] * 510

        min = self.timesteps[t]['min']
        max = self.timesteps[t]['max']

        values = {}
        # for all the paths
        for i in self.pathselect:
            # if the path has the current time, add it
            if self.paths[i][0][1] <= t and self.paths[i][-1][1] >= t:
                j = t - self.paths[i][0][1]
                
                values[self.paths[i][j]] = None

        # add the clusters up
        for j in values.keys():
            c = self.timesteps[j[1]]['clusters'][j[0]]['o_timehistogram'][self.width / 2]

            c = adjustOneHistogram(c, min, max, max - min,\
                                       min, max, max - min, 510)

            total = map(lambda x: x[0] + x[1], zip(total, c))

        return total

    # def save a HSV transfer function
    def saveHSV(self, fn, c, a, minv, maxv):
        out = file(fn + '.tf', 'w')

        out.write('%f %f\n' % (minv - (maxv - minv) / 510.0, maxv + (maxv - minv) / 510.0))
        out.write('0.0 0.0 0.0 0.0\n')
        for i in range(0, 510):
            ci = hsv2rgb((c[i][0] + c[i+1][0]) * .5,\
                             (c[i][1] + c[i+1][1]) * .5,\
                             (c[i][2] + c[i+1][2]) * .5)
            ai = (a[i] + a[i+1]) * .5

            out.write('%e %e %e %e\n' % (ci[0] * ai, ci[1] * ai, ci[2] * ai, ai))
        out.write('0.0 0.0 0.0 0.0\n')
        out.close()
        
    # save the transfer function, red-blue HSV, static/static
    def saveHSVStaticStatic(self):
        fn = asksaveasfilename()
        if fn == None or len(self.pathselect) <= 0:
            return

        minv = header['min']
        maxv = header['max']

        c = cdf(self.accumulateAll())

        a = sineOpacity(c, float(self.opacityentry.get()), float(self.opowerentry.get()))
        c = map(lambda x: (240 * x / float(c[-1]), .8, .8), c)

        self.saveHSV(fn, c, a, minv, maxv)

    # save the transfer functions, red-blue HSV, dynamic/dynamic
    def saveHSVDynamicDynamic(self):
        fn = asksaveasfilename()
        if fn == None or len(self.pathselect) <= 0:
            return

        for t in range(0, self.ts):
            minv = self.timesteps[t]['min']
            maxv = self.timesteps[t]['max']

            c = cdf(self.accumulate(t))
 
            a = sineOpacity(c, float(self.opacityentry.get()), float(self.opowerentry.get()))
            c = map(lambda x: (240 * x / float(c[-1]), .8, .8), c)

            self.saveHSV((fn + '-%d') % (t * self.data['skip'] + self.data['start']), c, a, minv, maxv)

    # save the transfer functions, red-blue HSV, static/dynamic
    def saveHSVStaticDynamic(self):
        fn = asksaveasfilename()
        if fn == None or len(self.pathselect) <= 0:
            return
        
        minv = header['min']
        maxv = header['max']

        c = cdf(self.accumulateAll())
        c = map(lambda x: (240 * x / float(c[-1]), .8, .8), c)

        for t in range(0, self.ts):
            amin = self.timesteps[t]['min']
            amax = self.timesteps[t]['max']

            a = self.accumulate(t)
            a = cdf(adjustOneHistogram(a, minv, maxv, maxv - minv,\
                                           amin, amax, amax - amin, 510))
            a = sineOpacity(a, float(self.opacityentry.get()), float(self.opowerentry.get()))
            
            self.saveHSV((fn + '-%d') % (t * self.data['skip'] + self.data['start']), c, a, minv, maxv)

    # save the transfer functions, red-blue HSV, dynamic/static
    def saveHSVDynamicStatic(self):
        fn = asksaveasfilename()
        if fn == None or len(self.pathselect) <= 0:
            return

        minv = header['min']
        maxv = header['max']

        a = cdf(self.accumulateAll())
        a = sineOpacity(a, float(self.opacityentry.get()), float(self.opowerentry.get()))
        
        for t in range(0, self.ts):
            cmin = self.timesteps[t]['min']
            cmax = self.timesteps[t]['max']

            c = self.accumulate(t)
            c = cdf(adjustOneHistogram(c, minv, maxv, maxv - minv,\
                                           cmin, cmax, cmax - cmin, 510))
            c = map(lambda x: (240 * x / float(c[-1]), .8, .8), c)
            
            self.saveHSV((fn + '-%d') % (t * self.data['skip'] + self.data['start']), c, a, minv, maxv)
                    
    def saveSimpleStatic(self):
        fn = asksaveasfilename()
        if fn == None or len(self.pathselect) <= 0:
            return

        # the middle time step
        t = self.ts / 2
        minv = self.timesteps[t]['min']
        maxv = self.timesteps[t]['max']
        c = cdf([1.0] * 510)

        a = sineOpacity(c, float(self.opacityentry.get()), float(self.opowerentry.get()))
        c = map(lambda x: (240 * x / float(c[-1]), .8, .8), c)

        self.saveHSV(fn, c, a, minv, maxv)

    def saveSimpleDynamic(self):
        fn = asksaveasfilename()
        if fn == None or len(self.pathselect) <= 0:
            return

        for t in range(0, self.ts):
            minv = self.timesteps[t]['min']
            maxv = self.timesteps[t]['max']
            c = cdf([1.0] * 510)

            a = sineOpacity(c, float(self.opacityentry.get()), float(self.opowerentry.get()))
            c = map(lambda x: (240 * x / float(c[-1]), .8, .8), c)

            self.saveHSV((fn + '-%d') % (t * self.data['skip'] + self.data['start']), c, a, minv, maxv)

    # Displays time histogram images per cluster
    def imageDisplay(self):
        xc = self.THXGAP
        c = 0
        for i in self.images:
            yc = self.THYGAP
            d = 0
            for j in i:
                self.canvas.create_image(xc, yc, image=j, anchor=NW)
                self.canvas.create_rectangle(xc - 1, yc - 1, xc + self.THWIDTH + 1, yc + self.THHEIGHT + 1, outline='red', tag='n%d,%d' % (d, c))

                yc = yc + self.THHEIGHT + self.THYGAP
                d = d + 1
            # write the time step text
            yc = self.maxclusters * (self.THHEIGHT + self.THYGAP) + self.THYGAP
            self.canvas.create_text(xc, yc, anchor=NW, fill='red', text='%d' % (c * self.data['skip'] + self.data['start']))

            xc = xc + self.THWIDTH + self.THXGAP
            c = c + 1

    # displays paths in the graph
    def pathDisplay(self):
        c = 0
        for i in self.paths:
            self.listbox.insert(END, "%d to %d\tlen: %d\tmin: %f\tavg: %f\tmul: %f" % (i[0][1] * self.data['skip'] + self.data['start'], i[-1][1] * self.data['skip'] + self.data['start'], self.metrics[c]['count'], self.metrics[c]['min'], self.metrics[c]['avg'], self.metrics[c]['mul']))
            for j in range(0, len(i) - 1):
                a = i[j]
                b = i[j + 1]

                ax = (a[1] + 1) * (self.THWIDTH + self.THXGAP)
                ay = a[0] * (self.THHEIGHT + self.THYGAP) + self.THYGAP + self.THHEIGHT / 2
                bx = b[1] * (self.THWIDTH + self.THXGAP) + self.THXGAP
                by = b[0] * (self.THHEIGHT + self.THYGAP) + self.THYGAP + self.THHEIGHT / 2

                self.canvas.create_line(ax, ay, bx, by, fill='red', tag='e%d' % c)
                
            c = c + 1

    # display single paths at the bottom
    def singlePathDisplay(self):
        yc = self.THYGAP * 2 + self.maxclusters * (self.THYGAP + self.THHEIGHT)
        p = 0
        for i in self.paths:
            q = 0
            for j in range(0, len(i)):
                a = i[j]

                ax = (a[1] + 1) * (self.THWIDTH + self.THXGAP)
                ay = yc + self.THYGAP + self.THHEIGHT / 2


                if j < len(i) - 1:
                    b = i[j + 1]
                    bx = b[1] * (self.THWIDTH + self.THXGAP) + self.THXGAP
                    by = yc + self.THYGAP + self.THHEIGHT / 2
                    
                    self.canvas.create_line(ax, ay, bx, by, fill='red', tag='e%d' % p)

                ax = ax - self.THWIDTH
                ay = ay - self.THHEIGHT / 2

                self.canvas.create_image(ax, ay, image=self.images[a[1]][a[0]], anchor=NW)
                self.canvas.create_rectangle(ax - 1, ay - 1, ax + self.THWIDTH + 1, ay + self.THHEIGHT + 1, outline='red', tag='n%d,%d' % (a[0], a[1]))

                q = q + 1

            ax = self.THXGAP + (self.THWIDTH + self.THXGAP) * self.ts
            ay = yc + self.THYGAP
            bx = ax + self.ts * self.PWIDTH
            by = ay + self.THHEIGHT

            self.canvas.create_image(ax, ay, image=self.pathimages[p], anchor=NW)
            self.canvas.create_rectangle(ax - 1, ay - 1, bx + 1, by + 1, outline='red')

            p = p + 1
            yc = yc + self.THYGAP + self.THHEIGHT

    # creates an image from a time histogram
    def imageFromTimeHisto(self, th):
        scale = max(map(max, th))
        if scale > 0:
            scale = 256.0 / log(scale)
        f = NamedTemporaryFile(suffix='.pgm')

        f.write('P5\n')
        f.write('%d %d\n' % (len(th) * self.PWIDTH, len(th[0])))
        f.write('255\n')

        for a in range(0, len(th[0])):
            for b in range(0, len(th)):
                v = 0
                if th[b][a] > 0:
                    v = int(log(th[b][a]) * scale)
                if v > 255:
                    v = 255
                if v < 0:
                    v = 0
                f.write(('%c' * self.PWIDTH) % tuple([v] * self.PWIDTH))

        f.flush()
        i = PhotoImage(file=f.name)
        f.close()
        return i

    # creates time histogram images for each cluster
    def imageCreate(self):
        self.images = []
        for i in self.timesteps:
            series = []

            for j in i['clusters']:
                series.append(self.imageFromTimeHisto(j['timehistogram']))
            
            self.images.append(series)

    # creates time histogram images for each path
    def pathImageCreate(self):
        self.pathimages = []
        for i in self.paths:
            th = []

            mint = min(map(lambda x: x[1], i))
            maxt = max(map(lambda x: x[1], i))

            for j in range(0, mint):
                th.append([0] * self.height)

            for j in range(0, len(i)):
                c = i[j][0]
                t = i[j][1]

                th.append(self.timesteps[t]['clusters'][c]['timehistogram'][self.width / 2])

            for j in range(maxt, self.ts):
                th.append([0] * self.height)

            self.pathimages.append(self.imageFromTimeHisto(th))

    # polls the listbox for selection changes
    def lbPoll(self):
        now = self.listbox.curselection()
        if now != self.lbselect:
            self.lbSelect(now)
            self.lbselect = now
        self.listbox.after(250, self.lbPoll)

    # updates the display when the listbox changes
    def lbSelect(self, now):
        pathselect = map(int, self.listbox.curselection())
        nodeselect = {}
        duplicates = {}
        for i in pathselect:
            for j in self.paths[i]:
                if j in nodeselect:
                    duplicates[j] = None
                else:
                    nodeselect[j] = None
                

        removededges = filter(lambda x: x not in pathselect, self.pathselect)
        removednodes = filter(lambda x: x not in nodeselect, self.nodeselect)

        options = {'fill' : 'red'}
        for i in removededges:
            self.canvas.itemconfig('e%d' % i, options)
        options = {'outline' : 'red'}
        for i in removednodes:
            self.canvas.itemconfig('n%d,%d' % i, options)

        options = {'fill' : 'white'}
        for i in pathselect:
            self.canvas.itemconfig('e%d' % i, options)
        options = {'outline' : 'white'}
        for i in nodeselect:
            self.canvas.itemconfig('n%d,%d' % i, options)

        options = {'outline' : 'yellow'}
        for i in duplicates:
            self.canvas.itemconfig('n%d,%d' % i, options)

        self.pathselect = pathselect
        self.nodeselect = nodeselect

#
# start the GUI
#

if len(argv) != 2:
    stderr.write('usage: %s input_sequence_file')

    stderr.write('required inputs:\n')
    stderr.write('input_sequence_file: sequenced data in python gzip pickled format created by the sequencer\n')

# read the data
f = GzipFile(argv[1], 'rb')
u = Unpickler(f)

data = u.load()
header = u.load()
timesteps = u.load()
graph = u.load()
weights = u.load()
paths = u.load()
metrics = u.load()

f.close()

# start it up
root = Tk()
main = GUI(root, data, header, timesteps, graph, weights, paths, metrics)
root.mainloop()
